# built-in
import ast
import re
from functools import partial
from importlib import import_module
from pathlib import Path
from typing import Dict, List


REX_CODE = re.compile(r'^[A-Z]{1,5}[0-9]{1,5}$')
ALIASES = {
    'flake8_bugbear': 'bugbear',
    'flake8_logging_format': 'logging_format',
}


class CollectStrings(ast.NodeVisitor):
    _strings: List[str]

    def visit_Str(self, node):
        self._strings.append(node.s)


def get_messages(code: str, content: str) -> Dict[str, str]:
    root = ast.parse(content)
    CollectStrings._strings = []
    collector = CollectStrings()
    collector.visit(root)

    messages = dict()
    for message in collector._strings:
        message_code, _, message_text = message.partition(' ')
        if not message_text:
            continue
        message_code = message_code.rstrip(':')
        if not REX_CODE.match(message_code):
            continue
        if code and not message_code.startswith(code):
            continue
        messages[message_code] = message_text
    return messages


def extract_default(name: str) -> Dict[str, str]:
    module = import_module(name)
    if module.__file__ is None:
        raise NotImplementedError(f'Unable to parse {name} module')
    content = Path(module.__file__).read_text()
    return get_messages(code='', content=content)


def extract(name) -> Dict[str, str]:
    name = name.replace('-', '_')
    name = ALIASES.get(name, name)
    function_name = 'extract_' + name

    # use ad-hoc extractor if available
    if function_name in globals():
        return globals()[function_name]()

    # try to extract by default algorithm
    return extract_default(name)


# -- EXTRACTORS WITH A CUSTOM PATH -- #


def extract_flake8_absolute_import() -> Dict[str, str]:
    return extract_default(name='flake8_absolute_import.core')


def extract_flake8_aaa() -> Dict[str, str]:
    return extract_default(name='flake8_aaa.line_markers')


def extract_flake8_cognitive_complexity() -> Dict[str, str]:
    return extract_default(name='flake8_cognitive_complexity.checker')


def extract_flake8_variables_names() -> Dict[str, str]:
    return extract_default(name='flake8_variables_names.checker')


def extract_logging_format() -> Dict[str, str]:
    return extract_default(name='logging_format.violations')


def extract_flake8_sql() -> Dict[str, str]:
    return extract_default(name='flake8_sql.linter')


def extract_flake8_requirements() -> Dict[str, str]:
    return extract_default(name='flake8_requirements.checker')


def extract_flake8_expression_complexity() -> Dict[str, str]:
    return extract_default(name='flake8_expression_complexity.checker')


def extract_flake8_use_fstring() -> Dict[str, str]:
    codes = dict()
    codes.update(extract_default(name='flake8_use_fstring.format'))
    codes.update(extract_default(name='flake8_use_fstring.percent'))
    # https://github.com/MichaelKim0407/flake8-use-fstring/pull/2
    try:
        codes.update(extract_default(name='flake8_use_fstring.prefix'))
    except ImportError:
        pass
    return codes


def extract_flake8_functions() -> Dict[str, str]:
    codes = dict()
    codes.update(extract_default('flake8_functions.checker'))
    try:
        codes.update(extract_default('flake8_functions.function_arguments_amount'))
        codes.update(extract_default('flake8_functions.function_lenght'))
        codes.update(extract_default('flake8_functions.function_purity'))
    except ImportError:
        pass
    return codes


# -- HARDCODED CODES -- #


def extract_flake8_spellcheck() -> Dict[str, str]:
    return {
        'SC100': 'Spelling error in comments',
        'SC200': 'Spelling error in name',
    }


def extract_flake8_import_order() -> Dict[str, str]:
    return {
        'I666': 'Import statement mixes groups.',
        'I100': 'Import statements are in the wrong order.',
        'I101': 'Imported names are in the wrong order.',
        'I201': 'Missing newline between import groups.',
        'I202': 'Additional newline in a group of imports.',
    }


def extract_flake8_black() -> Dict[str, str]:
    # external
    from flake8_black import black_prefix

    return {
        black_prefix + '901': 'Invalid input',
        black_prefix + '997': 'Invalid TOML file',
        black_prefix + '999': 'Unexpected exception',
    }


def extract_flake8_alfred() -> Dict[str, str]:
    return {'B1': 'banned symbol'}


def extract_flake8_eradicate() -> Dict[str, str]:
    return {'E800': 'Found commented out code: {0}'}


# -- AD-HOC EXTRACTORS -- #


def extract_flake8_commas() -> Dict[str, str]:
    # external
    from flake8_commas._base import ERRORS

    return dict(ERRORS.values())


def extract_flake8_debugger() -> Dict[str, str]:
    # external
    from flake8_debugger import DEBUGGER_ERROR_CODE
    return {DEBUGGER_ERROR_CODE: 'trace found'}


def extract_flake8_mutable() -> Dict[str, str]:
    # external
    from mutable_defaults import MutableDefaultChecker

    return {MutableDefaultChecker._code: MutableDefaultChecker._error_tmpl}


def extract_flake8_fixme() -> Dict[str, str]:
    # external
    from flake8_fixme import WORD_CODES

    return {code: 'fixme found ({})'.format(word) for word, code in WORD_CODES.items()}


def extract_pep8_naming() -> Dict[str, str]:
    # external
    import pep8ext_naming

    codes = dict()
    for checker_name in dir(pep8ext_naming):
        if not checker_name.endswith('Check'):
            continue
        checker = getattr(pep8ext_naming, checker_name)
        for code, message in checker.__dict__.items():
            if code[0] == 'N':
                codes[code] = message
    return codes


def extract_flake8_pyi() -> Dict[str, str]:
    # external
    import pyi
    codes = dict()
    for name, value in vars(pyi).items():
        if name.startswith('Y0'):
            codes[name] = value
    return codes


def extract_flake8_pytest_style() -> Dict[str, str]:
    # external
    from flake8_pytest_style import errors
    codes = dict()
    for error in vars(errors).values():
        if error is errors.Error:
            continue
        if not isinstance(error, type):
            continue
        if not issubclass(error, errors.Error):
            continue
        codes[error.code] = error.message  # type: ignore
    return codes


def extract_flake8_annotations_complexity() -> Dict[str, str]:
    _error_message_templates = ['TAE002 too complex annotation ({0} > {1})',
                                'TAE003 too long annotation ({0} > {1})']
    codes = dict()
    for _error_message_template in _error_message_templates:
        code, message = _error_message_template.split(' ', maxsplit=1)
        codes[code] = message
    return codes


def extract_flake8_future_import() -> Dict[str, str]:
    # external
    from flake8_future_import import ALL_FEATURES
    codes = dict()
    tmpl = 'FI{}'
    for feature in ALL_FEATURES:
        code = tmpl.format(10 + feature.index)
        codes[code] = '__future__ import "{}" missing'.format(feature.name)
        code = tmpl.format(50 + feature.index)
        codes[code] = '__future__ import "{}" present'.format(feature.name)
    codes[tmpl.format(90)] = '__future__ import does not exist'
    return codes


def extract_flake8_string_format() -> Dict[str, str]:
    # external
    from flake8_string_format import StringFormatChecker

    return {'P{}'.format(c): m for c, m in StringFormatChecker.ERRORS.items()}


def extract_flake8_bandit() -> Dict[str, str]:
    # external
    from bandit.core.extension_loader import MANAGER

    codes = dict()
    for blacklist in MANAGER.blacklist.values():
        for check in blacklist:
            code = check['id'].replace('B', 'S')
            codes[code] = check['message']
    for plugin in MANAGER.plugins:
        code = plugin.plugin._test_id.replace('B', 'S')
        codes[code] = plugin.name.replace('_', ' ')
    return codes


def extract_pylint() -> Dict[str, str]:
    # external
    import pylint.checkers
    try:
        # external
        from pylint.lint import MSGS
    except ImportError:
        # external
        from pylint.lint.pylinter import MSGS

    codes = dict()
    for code, (msg, alias, *_) in MSGS.items():
        if msg in ('%s', '%s: %s'):
            msg = alias.replace('-', ' ')
        codes[code] = msg.replace('\n', ' ')

    for path in Path(pylint.checkers.__path__[0]).iterdir():
        module = import_module('pylint.checkers.' + path.stem)
        for class_name in dir(module):
            cls = getattr(module, class_name, None)
            msgs = getattr(cls, 'msgs', None)
            if not msgs:
                continue
            for code, (msg, alias, *_) in msgs.items():
                if msg in ('%s', '%s: %s'):
                    msg = alias.replace('-', ' ')
                codes[code] = msg.replace('\n', ' ')
    return codes


def extract_pyflakes() -> Dict[str, str]:
    # external
    from flake8.plugins.pyflakes import FLAKE8_PYFLAKES_CODES
    from pyflakes import messages

    codes = dict()
    for class_name, code in FLAKE8_PYFLAKES_CODES.items():
        codes[code] = getattr(messages, class_name).message
    return codes


def extract_flake8_rst_docstrings() -> Dict[str, str]:
    # external
    from flake8_rst_docstrings import code_mappings_by_level

    codes = dict()
    for level, codes_mapping in code_mappings_by_level.items():
        for message, number in codes_mapping.items():
            code = 'RST{}{:02d}'.format(level, number)
            codes[code] = message
    return codes


def extract_flake8_django() -> Dict[str, str]:
    # external
    import flake8_django.checkers

    codes = dict()
    for path in Path(flake8_django.checkers.__path__[0]).iterdir():
        module = import_module('flake8_django.checkers.' + path.stem)
        for class_name in dir(module):
            cls = getattr(module, class_name)
            if not hasattr(cls, 'code'):
                continue
            if '0' not in cls.__name__:
                continue
            codes[cls.__name__] = cls.description
    return codes


def extract_flake8_scrapy() -> Dict[str, str]:
    # external
    from flake8_scrapy import ScrapyStyleIssueFinder

    codes = dict()
    for finders in ScrapyStyleIssueFinder().finders.values():
        for finder in finders:
            codes[finder.msg_code] = finder.msg_info
    return codes


def extract_flake8_executable() -> Dict[str, str]:
    # external
    import flake8_executable

    path = Path(flake8_executable.__file__)
    content = path.read_text()
    codes = dict()
    for code, msg in re.findall(r"'(EXE00\d)', '(.*)'", content):
        codes[code] = msg
    return codes


def extract_flake8_docstrings() -> Dict[str, str]:
    # external
    from pydocstyle.violations import ErrorRegistry

    codes = dict()
    for group in ErrorRegistry.groups:
        for error in group.errors:
            codes[error.code] = error.short_desc
    return codes


def extract_dlint() -> Dict[str, str]:
    # external
    from dlint.linters import ALL

    codes = dict()
    for linter in ALL:
        code, msg = linter._error_tmpl.split(' ', maxsplit=1)
        codes[code] = msg
    return codes


def extract_flake8_mock() -> Dict[str, str]:
    # external
    from flake8_mock import ERROR_MESSAGE, MOCK_ERROR_CODE

    message = ERROR_MESSAGE.split(' ', maxsplit=1)[1]
    return {MOCK_ERROR_CODE: message}


def extract_flake8_pytest() -> Dict[str, str]:
    # external
    from flake8_pytest import PYTEST_ERROR_CODE, PYTEST_ERROR_MESSAGE

    return {PYTEST_ERROR_CODE: PYTEST_ERROR_MESSAGE}


def extract_wemake_python_styleguide() -> Dict[str, str]:
    # external
    from wemake_python_styleguide import violations

    codes = dict()
    for path in Path(violations.__path__[0]).iterdir():
        module = import_module('wemake_python_styleguide.violations.' + path.stem)
        for checker_name in dir(module):
            if not checker_name.endswith('Violation'):
                continue
            checker = getattr(module, checker_name)
            if not hasattr(checker, 'code'):
                continue
            code = 'WPS' + str(checker.code).zfill(3)
            codes[code] = checker.error_template
    return codes


def extract_flake8_pie() -> Dict[str, str]:
    # built-in
    from inspect import getsource

    # external
    import flake8_pie

    codes = dict()
    for path in Path(flake8_pie.__path__[0]).iterdir():
        module = import_module('flake8_pie.' + path.stem)
        for code in dir(module):
            if not code.startswith('PIE'):
                continue
            error = getattr(module, code)
            if isinstance(error, partial):
                code_message = error.keywords['message']
            else:
                fn_source = getsource(error)
                code_message = fn_source.split('message=')[1].strip('\n (),\'"')
            codes[code] = code_message.split(': ', maxsplit=1)[1]
    return codes



try:
    # built-in
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    # external
    import importlib_metadata as metadata  # type: ignore

__version__ = metadata.version('flakeheaven')



# built-in
from pathlib import Path
from types import MappingProxyType
from typing import List

# app
from ._base import BaseParser


class YAMLParser(BaseParser):
    ignore = MappingProxyType({
        'pycodestyle': ('E302', 'E303', 'E305', 'E402'),
    })

    @classmethod
    def parse(cls, path: Path) -> List[str]:
        if not path.name.startswith(('test-', 'test_')):
            return []
        with path.open(encoding='utf8') as stream:
            return cls._pytest_mypy_plugins(stream)

    @staticmethod
    def _pytest_mypy_plugins(stream) -> List[str]:
        """Parse pytest-mypy-plugins tests

        https://github.com/typeddjango/pytest-mypy-plugins
        """
        code_block = False
        code_found = False
        indent = None
        lines = []
        for line in stream:
            if not line.strip():
                lines.append('\n')
                continue

            if not code_block:
                # start of code block
                if line.lstrip().startswith('main: |'):
                    code_block = True
                    indent = None
                # ignore not code blocks
                lines.append('# ' + line.lstrip())
                continue

            # end of code block
            current_indent = len(line) - len(line.lstrip())
            if indent is not None and current_indent < indent:
                code_block = False
                lines.append('# ' + line.lstrip())
                continue

            # wite line from a code block
            if indent is None:
                indent = current_indent
            lines.append(line[indent:])
            code_found = True

        if not code_found:
            return []
        # Replace the first line (it can't be an actual code) by a mock for `reveal_type`.
        lines[0] = 'reveal_type = lambda x: x  # noqa\n'
        return lines



# built-in
from collections import namedtuple

# external
from flake8.style_guide import Violation


# the same as in flake8, but with some additional fields
_Violation = namedtuple(
    '_Violation',
    [
        'code',
        'filename',
        'line_number',
        'column_number',
        'text',
        'physical_line',

        # added fields
        'plugin',
    ],
)


class FlakeHeavenViolation(_Violation):
    """Patched flake8.style_guide.Violation

    We can't just inherit because Violation is a namedtuple,
    and we can't add new fields.
    """

    def is_inline_ignored(self, disable_noqa):
        return Violation.is_inline_ignored(self, disable_noqa)

    def is_in(self, diff):
        return Violation.is_in(self, diff)



# built-in
from pathlib import Path
from types import MappingProxyType
from typing import List, Mapping, Tuple


class BaseParser:
    ignore: Mapping[str, Tuple[str, ...]] = MappingProxyType({})

    @staticmethod
    def parse(path: Path) -> List[str]:
        raise NotImplementedError



# built-in
from pathlib import Path
from typing import List, Type

# external
from flake8.processor import FileProcessor

# app
from ..parsers import PARSERS, BaseParser, PythonParser


class FlakeHeavenProcessor(FileProcessor):
    parser: Type[BaseParser] = PythonParser

    def read_lines_from_filename(self) -> List[str]:
        """Read the lines for a file."""
        path = Path(self.filename)
        self.parser = PARSERS.get(path.suffix, PythonParser)
        return self.parser.parse(path=path)



# built-in
from pathlib import Path

# external
from flake8.formatting.base import BaseFormatter

# app
from .._logic import make_baseline


class BaseLineFormatter(BaseFormatter):
    def format(self, error):
        filename = error.filename
        return make_baseline(
            path=Path(filename),
            code=error.code,
            line=error.line_number,
            context=error.physical_line,
        )



# built-in
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

# external
from flake8.checker import FileChecker, Manager
from flake8.utils import filenames_from, fnmatch

# app
from .._logic import (
    Snapshot, check_include, get_exceptions, get_plugin_name,
    get_plugin_rules, make_baseline, prepare_cache,
)
from ._processor import FlakeHeavenProcessor


DEFAULT_PLUGIN = 'pycodestyle'


class Result(NamedTuple):
    plugin_name: str
    error_code: str
    line_number: int
    column: int
    text: str
    line: str


def is_relative_to(path: Path, maybe_parent: Path):
    try:
        path.relative_to(maybe_parent)
    except ValueError:
        return False
    else:
        return True


class FlakeHeavenCheckersManager(Manager):
    """
    Patched flake8.checker.Manager to provide `plugins` support
    """

    def __init__(self, baseline: Optional[str], **kwargs):
        self.baseline = set()
        self.relative = kwargs.pop('relative', False)
        if baseline:
            with open(baseline) as stream:
                self.baseline = {line.strip() for line in stream}
            self.root_path = Path(baseline).resolve().parent
        super().__init__(**kwargs)

    def make_checkers(self, paths: List[str] = None) -> None:
        """
        Reloaded checkers generator to provide one checker per file per rule.
        Original `make_checkers` provides checker per file with all rules mixed.
        It makes difficult to filter checks by codes after all.
        """
        if paths is None:
            paths = self.arguments
        if not paths:
            paths = ['.']
        prepare_cache()

        # `checkers` is list of checks to run (and then cache)
        # check is a combination of plugin and file.
        self.checkers = []
        # `snapshots` is the list of checks that have cache and should not be run
        self.snapshots = []
        for argument in paths:
            for filename in filenames_from(argument, self.is_path_excluded):
                # get checks list
                selected_checks: Dict[str, List[Dict[str, Any]]]
                selected_checks = dict(
                    ast_plugins=[],
                    logical_line_plugins=[],
                    physical_line_plugins=[],
                )
                has_checks = False
                for check_type, checks in self.checks.to_dictionary().items():
                    for check in checks:
                        should_process = self._should_process(
                            argument=argument,
                            filename=filename,
                            check_type=check_type,
                            check=check,
                        )
                        if not should_process:
                            continue
                        selected_checks[check_type].append(check)
                        has_checks = True

                # Create checker with selected checks
                if not has_checks:
                    continue
                checker = FlakeHeavenFileChecker(
                    filename=filename,
                    checks=selected_checks,
                    options=self.options,
                )
                # ignore files with top-level `# flake8: noqa`
                if not checker.should_process:
                    continue
                checker.snapshot = Snapshot.create(
                    checker=checker,
                    options=self.options,
                )
                if checker.snapshot.exists():
                    self.snapshots.append(checker)
                    continue
                self.checkers.append(checker)

    def _should_process(
        self, argument: str, filename: str, check_type: str, check: Dict[str, Any],
    ) -> bool:
        # do not run plugins without rules specified
        plugin_name = get_plugin_name(check)
        rules = self._get_rules(plugin_name=plugin_name, filename=filename)
        if not rules or set(rules) == {'-*'}:
            return False

        if filename == '-':
            return True
        if fnmatch(filename=filename, patterns=self.options.filename):
            return True

        if self.options._running_from_vcs:
            return False
        if self.options.diff:
            return False
        return argument == filename

    def _get_rules(self, plugin_name: str, filename: str) -> List[str]:
        rules = get_plugin_rules(
            plugin_name=plugin_name,
            plugins=self.options.plugins,
        )
        exceptions = get_exceptions(
            path=filename,
            exceptions=self.options.exceptions,
        )
        if exceptions:
            rules = rules.copy()
            rules += get_plugin_rules(
                plugin_name=plugin_name,
                plugins=exceptions,
            )
        return rules

    def is_path_excluded(self, filename: str) -> bool:
        """Patched `is_path_excluded`.

        We patch it to exclude files not specified explicitly.
        It is helpful when you want to run flakeheaven with `--diff`
        and explicitly passed paths at the same time.

        Run flakeheaven only on changes in example.py:

            git diff | flakeheaven lint --diff example.py
        """
        if filename == '-':
            filename = self.options.stdin_display_name
        if super().is_path_excluded(path=filename):
            return True

        # skip subpath check below for stdin
        arguments = set(self.arguments) - {'.', '-', 'stdin'}
        if not arguments:
            return False

        # include file only if it is a subpath of
        # an explicitly specified CLI argument.
        path = Path(filename).absolute()
        parents = path.parents
        for base_filename in self.arguments:
            base_path = Path(base_filename).absolute()
            if base_path == path:
                return False
            if base_path in parents:
                return False
        return True

    def report(self) -> Tuple[int, int]:
        """Reloaded report generation to filter out excluded error codes.

        + use checker.filename as path instead of checker.display_name
        + pass checker into `_handle_results` to get plugin name.
        """
        # self.run_serial()
        results_reported = results_found = 0
        for checker in self.checkers + self.snapshots:
            if not checker.results and not checker.snapshot.exists():
                continue

            # get results either from cache or actual run
            if checker.snapshot.exists():
                all_results = checker.snapshot.results
            else:
                all_results = sorted(
                    checker.results, key=lambda tup: (tup[1], tup[2]))
                checker.snapshot.dump(all_results)

            # group results by plugin name
            grouped_results = defaultdict(list)
            for result in all_results:
                if type(result) is not Result:
                    if len(result) == 6:
                        # cache entry is deserialized into list
                        result = Result(*result)
                    else:
                        # flake8 sets custom error codes in a few places
                        # where we didn't set `_processed_plugin`
                        result = Result(DEFAULT_PLUGIN, *result)
                grouped_results[result.plugin_name].append(result)

            # get filename
            filename = checker.filename
            if filename is None or filename == '-':
                filename = self.options.stdin_display_name or 'stdin'

            with self.style_guide.processing_file(filename):
                ignored = checker.processor.parser.ignore
                for plugin_name, results in sorted(grouped_results.items()):
                    results_reported += self._handle_results(
                        filename=filename,
                        results=results,
                        plugin_name=plugin_name,
                        ignored_codes=ignored.get(plugin_name, ()),
                    )
            results_found += len(all_results)
        return (results_found, results_reported)

    def _handle_results(
        self, filename: str, results: list, plugin_name: str, ignored_codes: Tuple[str, ...],
    ) -> int:
        rules = self._get_rules(plugin_name=plugin_name, filename=filename)
        reported_results_count = 0
        for result in results:
            # Some codes are ignored for a specific parser.
            # For example, lack of blank lines for YAML parser.
            if result.error_code in ignored_codes:
                continue

            # skip baselined errors
            if self.baseline:
                _path = Path(filename)
                if self.relative and is_relative_to(_path, self.root_path):
                    _path = _path.relative_to(self.root_path)
                digest = make_baseline(
                    path=_path,
                    context=result.line,
                    code=result.error_code,
                    line=result.line_number,
                )
                if digest in self.baseline:
                    continue

            # skip explicitly excluded codes
            if not check_include(code=result.error_code, rules=rules):
                continue

            # report
            reported_results_count += self.style_guide.handle_error(
                code=result.error_code,
                filename=filename,
                line_number=result.line_number,
                column_number=result.column,
                text=result.text,
                physical_line=result.line,
                plugin=plugin_name,
            )
        return reported_results_count


class FlakeHeavenFileChecker(FileChecker):
    """
    A little bit patched FileChecker to support `--safe`
    """
    snapshot: Snapshot
    _processed_plugin: str = DEFAULT_PLUGIN

    def _make_processor(self) -> Optional[FlakeHeavenProcessor]:
        try:
            return FlakeHeavenProcessor(self.filename, self.options)
        except IOError as e:
            message = '{0}: {1}'.format(type(e).__name__, e)
            self.report('E902', 0, 0, message)
            return None

    def run_checks(self) -> Tuple[str, List[Result], Dict[str, Any]]:
        if not self.processor:
            return self.filename, self.results, self.statistics
        if not self.processor.lines:
            return self.filename, self.results, self.statistics
        try:
            return super().run_checks()
        except Exception as exc:
            if self.options.safe:
                message = '{0}: {1}'.format(type(exc).__name__, exc)
                self.report('E902', 0, 0, message)
            else:
                raise
        return self.filename, self.results, self.statistics

    def run_check(self, plugin: Dict[str, Any], **arguments):
        self._processed_plugin = get_plugin_name(plugin)
        return super().run_check(plugin=plugin, **arguments)

    def report(self, error_code: Optional[str], line_number: int, column: int, text: str) -> str:
        """
        Copy-pasted `report` method to store `Result` in `self.results` instead of tuple
        and provide `plugin_name`.
        """
        if error_code is None:
            error_code, text = text.split(' ', 1)

        # If we're recovering from a problem in _make_processor, we will not
        # have this attribute.
        if hasattr(self, 'processor'):
            line = self.processor.noqa_line_for(line_number)
        else:
            line = None

        self.results.append(Result(
            plugin_name=self._processed_plugin,
            error_code=error_code,
            line_number=line_number,
            column=column,
            text=text,
            line=line,
        ))
        return error_code



# built-in
import tokenize
from pathlib import Path
from typing import List

# app
from ._base import BaseParser


class PythonParser(BaseParser):
    @staticmethod
    def parse(path: Path) -> List[str]:
        try:
            with tokenize.open(str(path)) as fd:
                return fd.readlines()
        except (SyntaxError, UnicodeError):
            with open(str(path), encoding='utf8') as fd:
                return fd.readlines()



# built-in
import json
from pathlib import Path

# external
from flake8.formatting.base import BaseFormatter

# app
from .._logic import make_baseline


class GitlabFormatter(BaseFormatter):
    error_format = '{code} {text}'

    def start(self):
        super().start()
        self._write('[')
        self.newline = ''
        self._first_line = True

    def stop(self):
        self._write('\n]\n')
        super().stop()

    def handle(self, error):
        # redefined to never output source
        line = self.format(error)
        self._write(line)

    def format(self, error):
        filename = error.filename
        digest = make_baseline(
            path=Path(filename),
            code=error.code,
            line=error.line_number,
            context=error.physical_line,
        )
        text = self.error_format.format(code=error.code, text=error.text)
        # docs.gitlab.com/ee/user/project/merge_requests/code_quality.html
        result = json.dumps(dict(
            description=text,
            fingerprint=digest,
            severity='info',
            location={
                'path': filename,
                'lines': dict(begin=error.line_number),
            },
        ))
        if self._first_line:
            self._first_line = False
        else:
            result = ',\n' + result
        return result



# built-in
from types import MappingProxyType
from typing import Mapping, Type

# app
from ._base import BaseParser
from ._jupyter import JupyterParser
from ._markdown import MarkdownParser
from ._python import PythonParser
from ._rst import RSTParser
from ._yaml import YAMLParser


__all__ = [
    'BaseParser',
    'JupyterParser',
    'MarkdownParser',
    'PARSERS',
    'PythonParser',
    'RSTParser',
    'YAMLParser',
]


PARSERS: Mapping[str, Type[BaseParser]] = MappingProxyType({
    '.ipynb': JupyterParser,
    '.md': MarkdownParser,
    '.py': PythonParser,
    '.rst': RSTParser,
    '.rest': RSTParser,
    '.yaml': YAMLParser,
    '.yml': YAMLParser,
})



"""Flake8 wrapper to make it nice and configurable
"""

# app
from ._cli import entrypoint, flake8_entrypoint
from ._version import __version__


__all__ = ['entrypoint', 'flake8_entrypoint', '__version__']



# external
from flake8 import __version__ as flake8_version

# app
from .._constants import ExitCode
from .._logic import colored
from .._types import CommandResult
from .._version import __version__ as flakeheaven_version


def version_command(argv) -> CommandResult:
    """Show FlakeHeaven version.
    """
    print('FlakeHeaven', colored(flakeheaven_version, 'green'))
    print('Flake8   ', colored(flake8_version, 'green'))
    print('For plugins versions use', colored('flakeheaven plugins', 'green'))
    return ExitCode.OK, ''



# app
from .._constants import NAME, VERSION, ExitCode
from .._logic import get_installed, get_plugin_rules
from .._patched import FlakeHeavenApplication
from .._types import CommandResult


def missed_command(argv) -> CommandResult:
    """Show patterns from the config that has no matched plugin installed.
    """
    if argv and argv[0] == '--help':
        print(missed_command.__doc__)
        return ExitCode.OK, ''
    if argv:
        return ExitCode.TOO_MANY_ARGS, 'the command does not accept arguments'

    app = FlakeHeavenApplication(program=NAME, version=VERSION)
    installed_plugins = sorted(get_installed(app=app), key=lambda p: p['name'])
    if not installed_plugins:
        return ExitCode.NO_PLUGINS_INSTALLED, 'no plugins installed'

    count = 0
    for pattern in app.options.plugins:
        for plugin in installed_plugins:
            rules = get_plugin_rules(
                plugin_name=plugin['name'],
                plugins={pattern: ['+*']},
            )
            if rules:
                break
        else:
            print(pattern)
            count += 1
    return count, ''



# built-in
from types import SimpleNamespace

# app
from .._constants import NAME, VERSION
from .._patched import FlakeHeavenApplication
from .._types import CommandResult
from ..formatters import BaseLineFormatter


def baseline_command(argv) -> CommandResult:
    """Generate baseline that can be used later to ignore errors.
    """
    app = FlakeHeavenApplication(program=NAME, version=VERSION)
    app.formatter = BaseLineFormatter(SimpleNamespace(
        output_file=None,
        show_source=False,
    ))
    try:
        app.run(['--baseline', '', '--exit-zero'] + argv)
        app.exit()
    except SystemExit as exc:
        return int(exc.code), ''
    raise RuntimeError('unreachable')



# built-in
from collections import defaultdict

# external
from flake8.plugins.manager import Checkers, PluginManager


class MultiDict:
    def __init__(self):
        self._data = defaultdict(list)

    def get(self, name: str, default=None):
        items = self._data[name]
        if items:
            return items[0]
        return default

    def getlist(self, name: str, default=None) -> list:
        items = self._data[name]
        if items:
            return items
        return default

    def __getitem__(self, name):
        return self._data[name]

    def __setitem__(self, name, value):
        self._data[name].append(value)

    def items(self):
        for name, values in self._data.items():
            for value in values:
                yield name, value

    def values(self):
        for values in self._data.values():
            for value in values:
                yield value


class FlakeHeavenPluginManager(PluginManager):
    def __init__(self, namespace, local_plugins=None):
        self.namespace = namespace
        self.plugins = MultiDict()
        self.names = []
        self._load_local_plugins(local_plugins or [])
        self._load_entrypoint_plugins()

    def map(self, func, *args, **kwargs):
        for plugin in self.plugins.values():
            yield func(plugin, *args, **kwargs)

    def versions(self):
        plugins_seen = set()
        for plugin in self.plugins.values():
            plugin_name = plugin.plugin_name
            if plugin.plugin_name in plugins_seen:
                continue
            plugins_seen.add(plugin_name)
            yield (plugin_name, plugin.version)


class FlakeHeavenCheckers(Checkers):
    def __init__(self, local_plugins=None):
        self.manager = FlakeHeavenPluginManager(
            namespace=self.namespace,
            local_plugins=local_plugins,
        )
        self.plugins_loaded = False



# built-in
from functools import lru_cache

# external
from flake8.style_guide import StyleGuide, StyleGuideManager

# app
from ._violation import FlakeHeavenViolation


class FlakeHeavenStyleGuideManager(StyleGuideManager):
    def __init__(self, options, formatter, decider=None):
        """Initialize our StyleGuide.

        .. todo:: Add parameter documentation.
        """
        super().__init__(options, formatter, decider)
        self.default_style_guide = FlakeHeavenStyleGuide(
            options, formatter, self.stats, decider=decider,
        )
        self.style_guides = [self.default_style_guide]
        self.style_guides.extend(self.populate_style_guides_with(options))

    @lru_cache(maxsize=None)
    def style_guide_for(self, filename: str):
        """Patched styleguide finder to give priority to flakeheaven's stileguides
        """
        guides = sorted(
            (g for g in self.style_guides if g.applies_to(filename)),
            key=lambda g: len(g.filename or ''),
            reverse=True,
        )
        for guide in guides:
            if isinstance(guide, FlakeHeavenStyleGuide):
                return guide
        return guides[0]

    def handle_error(
        self,
        code: str,
        filename: str,
        line_number: int,
        column_number: int,
        text: str,
        plugin: str,
        physical_line: str = None,
    ):
        guide = self.style_guide_for(filename)
        params = dict(
            code=code,
            filename=filename,
            line_number=line_number,
            column_number=column_number,
            text=text,
            physical_line=physical_line,
        )
        if isinstance(guide, FlakeHeavenStyleGuide):
            params['plugin'] = plugin
        return guide.handle_error(**params)


class FlakeHeavenStyleGuide(StyleGuide):
    def handle_error(
        self,
        code: str,
        filename: str,
        line_number: int,
        column_number: int,
        text: str,
        plugin: str,
        physical_line: str = None,
    ):
        """
        This function copied as is, but:
        1. Violation is replaced by FlakeHeavenViolation
        2. `error_is_selected` dropped. If we get here, the error IS selected.
        """
        disable_noqa = self.options.disable_noqa
        if not column_number:
            column_number = 0
        error = FlakeHeavenViolation(
            code=code,
            filename=filename,
            line_number=line_number,
            column_number=column_number + 1,
            text=text,
            physical_line=physical_line,
            plugin=plugin,
        )
        is_not_inline_ignored = error.is_inline_ignored(disable_noqa) is False
        is_included_in_diff = error.is_in(self._parsed_diff)
        if is_not_inline_ignored and is_included_in_diff:
            self.formatter.handle(error)
            self.stats.record(error)
            return 1
        return 0



# built-in
from ast import AST
from tokenize import TokenInfo
from typing import Sequence


try:
    # external
    from pylint.__pkginfo__ import __version__ as version
    from pylint.lint import Run
    from pylint.reporters import BaseReporter
except ImportError:
    version = '0.0.0'
    Run = None
    BaseReporter = object


STDIN = 'stdin'


class Reporter(BaseReporter):
    def __init__(self):
        self.errors = []
        super().__init__()

    def _display(self, layout):
        pass

    def handle_message(self, msg):
        # ignore `invalid syntax` messages, it is already checked by `pycodestyle`
        if msg.msg_id == 'E0001':
            return
        self.errors.append(dict(
            row=msg.line,
            col=msg.column,
            text='{} {} ({})'.format(msg.msg_id, msg.msg or '', msg.symbol),
            code=msg.msg_id,
        ))


class PyLintChecker:
    name = 'pylint'
    version = version

    def __init__(self, tree: AST, file_tokens: Sequence[TokenInfo], filename: str = STDIN) -> None:
        self.tree = tree
        self.filename = filename
        self.file_tokens = file_tokens

    def run(self):
        # pylint is not installed, skip
        if Run is None:
            return

        reporter = Reporter()
        Run([self.filename], reporter=reporter, do_exit=False)
        for error in reporter.errors:
            yield error['row'], error['col'], error['text'], type(self)



# built-in
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict

# external
import toml
from flake8 import _EXTRA_VERBOSE, LOG as FLAKE8_LOG
from flake8.utils import normalize_paths


LOG = FLAKE8_LOG.getChild(__name__)


def read_config(*paths) -> Dict[str, Any]:
    config = dict()  # type: Dict[str, Any]
    for path in paths:
        if isinstance(path, Path):
            new_config = _read_local(path)
        elif path.startswith(('https://', 'http://')):
            new_config = _read_remote(path)
        elif Path(path).exists():
            new_config = _read_local(Path(path))
        else:
            new_config = _read_remote(path)
        LOG.log(_EXTRA_VERBOSE, 'CONFIG: incoming from `%s`:```%s```', path, new_config)
        config = _merge_configs(config, new_config)
        LOG.log(
            _EXTRA_VERBOSE, 'CONFIG: after merging from `%s`:```%s```', path, config,
        )
    return config


def _read_local(path: Path) -> Dict[str, Any]:
    with path.open('r') as stream:
        return _parse_config(stream.read())


def _read_remote(url: str) -> Dict[str, Any]:
    import urllib3  # isort: skip
    http = urllib3.PoolManager()
    response = http.request('GET', url)
    return _parse_config(response.data.decode())


def _merge_configs(*configs) -> Dict[str, Any]:
    config = defaultdict(dict)  # type: Dict[str, Any]
    for subconfig in configs:
        config.update(subconfig)

    for section in ('plugins', 'exceptions'):
        for subconfig in configs:
            if section in subconfig:
                config[section].update(subconfig[section])

    return dict(config)


def _parse_config(content: str) -> Dict[str, Any]:
    config = toml.loads(content).get('tool', {}).get('flakeheaven', {})
    config = dict(config)

    for section in ('plugins', 'exceptions'):
        if section in config:
            config[section] = dict(config[section])

    if 'base' in config:
        paths = config['base']
        if not isinstance(paths, list):
            paths = [paths]
        config = _merge_configs(read_config(*paths), config)

    if 'exclude' in config:
        config['exclude'] = normalize_paths(config['exclude'])

    return config



# built-in
from enum import Enum
from pathlib import Path
from types import MappingProxyType
from typing import List, Optional

# app
from ._base import BaseParser


class CodeType(Enum):
    PYTHON = 'python'
    PYCON = 'pycon'


CODE_TYPES = {
    'python': CodeType.PYTHON,
    'python3': CodeType.PYTHON,
    'py3': CodeType.PYTHON,
    'python2': CodeType.PYTHON,
    'py2': CodeType.PYTHON,

    'pycon': CodeType.PYCON,
}


class MarkdownParser(BaseParser):
    ignore = MappingProxyType({
        'pycodestyle': ('E302', 'E303', 'E305', 'E402'),
    })

    @classmethod
    def parse(cls, path: Path) -> List[str]:
        code_found = False
        code_type = None
        indent = None
        lines = []
        with path.open(encoding='utf8') as stream:
            for line in stream:
                # leave empty lines as-is
                if not line.strip():
                    lines.append('\n')
                    continue

                if code_type is None:
                    # detect code block start
                    new_code_type = cls._get_code_type(line=line)
                    if new_code_type:
                        code_type = new_code_type
                    # ignore markdown and code block starts
                    lines.append('# {}\n'.format(line[:40].strip()))
                    continue

                # detect code block end
                if line.strip() == '```':
                    code_type = None
                    lines.append('# {}\n'.format(line[:40].strip()))
                    indent = None
                    continue

                # For the first line of code check indentation.
                if indent is None:
                    indent = len(line) - len(line.lstrip())
                    if line.lstrip()[:4] == '>>> ':
                        code_type = CodeType.PYCON
                # Remove this identation from every line of the code block
                line = line[indent:]

                # remove repl chars and comment-out output
                if code_type == CodeType.PYCON:
                    if line.startswith('>>> ') or line.startswith('... '):
                        line = line[4:]
                    else:
                        line = '# ' + line

                # save code line as-is
                lines.append(line)
                code_found = True
        if not code_found:
            return []
        return lines

    @staticmethod
    def _get_code_type(line: str) -> Optional[CodeType]:
        line = line.lstrip()
        if line[:3] != '```':
            return None
        words = line[3:].lower().split()
        if not words:
            return None
        lang = words[0]
        return CODE_TYPES.get(lang, None)



# built-in
import json
from pathlib import Path
from typing import List

# app
from ._base import BaseParser


class JupyterParser(BaseParser):
    @staticmethod
    def parse(path: Path) -> List[str]:
        lines = []
        with path.open(encoding='utf8') as stream:
            notebook = json.load(stream)
            for cell in notebook['cells']:
                if cell['cell_type'] != 'code':
                    continue
                lines.append('\n')
                lines.append('# In [{}]:\n'.format(cell.get('execution_count', 0)))
                lines.extend(line.rstrip('\n') + '\n' for line in cell['source'])
        return lines



# built-in
from types import MappingProxyType

# app
from ._baseline import baseline_command
from ._code import code_command
from ._codes import codes_command
from ._config import config_command
from ._lint import lint_command
from ._missed import missed_command
from ._plugins import plugins_command
from ._version import version_command


__all__ = [
    'COMMANDS',

    'baseline_command',
    'code_command',
    'codes_command',
    'config_command',
    'lint_command',
    'missed_command',
    'plugins_command',
    'version_command',
]


COMMANDS = MappingProxyType({
    'baseline': baseline_command,
    'code': code_command,
    'codes': codes_command,
    'config': config_command,
    'lint': lint_command,
    'missed': missed_command,
    'plugins': plugins_command,
    '--version': version_command,
})



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flakeheaven',
 'flakeheaven._logic',
 'flakeheaven._patched',
 'flakeheaven.commands',
 'flakeheaven.formatters',
 'flakeheaven.parsers',
 'flakeheaven.plugins']

package_data = \
{'': ['*']}

install_requires = \
['colorama',
 'entrypoints',
 'flake8>=4.0.1,<5.0.0',
 'pygments',
 'toml',
 'urllib3']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0'],
 'docs': ['alabaster',
          'pygments-github-lexers',
          'sphinx',
          'myst-parser>=0.18.0,<0.19.0']}

entry_points = \
{'console_scripts': ['flake8heavened = flakeheaven:flake8_entrypoint',
                     'flakeheaven = flakeheaven:entrypoint'],
 'flake8.extension': ['pylint = flakeheaven.plugins:PyLintChecker'],
 'flake8.report': ['baseline = flakeheaven.formatters:BaseLineFormatter',
                   'colored = flakeheaven.formatters:ColoredFormatter',
                   'gitlab = flakeheaven.formatters:GitlabFormatter',
                   'grouped = flakeheaven.formatters:GroupedFormatter',
                   'json = flakeheaven.formatters:JSONFormatter',
                   'stat = flakeheaven.formatters:StatFormatter']}

setup_kwargs = {
    'name': 'flakeheaven',
    'version': '2.1.0',
    'description': 'FlakeHeaven is a [Flake8](https://gitlab.com/pycqa/flake8) wrapper to make it cool.',
    'long_description': '# FlakeHeaven\n\n[![License: MIT](https://img.shields.io/pypi/l/flakeheaven)](https://opensource.org/licenses/MIT)\n[![python versions](https://img.shields.io/pypi/pyversions/flakeheaven)](https://pypi.org/project/flakeheaven/)\n\n[![version](https://img.shields.io/pypi/v/flakeheaven)](https://pypi.org/project/flakeheaven/)\n[![conda](https://anaconda.org/conda-forge/flakeheaven/badges/version.svg)](https://anaconda.org/conda-forge/flakeheaven)\n[![Downloads](https://img.shields.io/pypi/dm/flakeheaven)](https://pypi.org/project/flakeheaven/)\n\n[![CI](https://github.com/flakeheaven/flakeheaven/actions/workflows/ci.yaml/badge.svg)](https://github.com/flakeheaven/flakeheaven/actions/workflows/ci.yaml)\n[![Docs](https://readthedocs.org/projects/flakeheaven/badge/?version=latest)](https://flakeheaven.readthedocs.io/en/latest/)\n\n\nFlakeHeaven is a [Flake8](https://gitlab.com/pycqa/flake8) wrapper to make it cool.\n\nThis project is a fork of [FlakeHell](https://github.com/life4/flakehell). FlakeHell and other forks of it such as\nflakehell/flakehell are [no longer maintained](https://github.com/flakehell/flakehell/issues/25) and do not work with Flake8 4.0.x.\n\nFlakeHeaven works with Flake8 4.0.1 or greater. This fork will be [maintained by the community](https://github.com/flakeheaven/flakeheaven/discussions/1) that developed the existing forks.\n\n+ [Lint md, rst, ipynb, and more](https://flakeheaven.readthedocs.io/en/latest/parsers.html).\n+ [Shareable and remote configs](https://flakeheaven.readthedocs.io/en/latest/config.html#base).\n+ [Legacy-friendly](https://flakeheaven.readthedocs.io/en/latest/commands/baseline.html): ability to get report only about new errors.\n+ Caching for much better performance.\n+ [Use only specified plugins](https://flakeheaven.readthedocs.io/en/latest/config.html#plugins), not everything installed.\n+ [Make output beautiful](https://flakeheaven.readthedocs.io/en/latest//formatters.html).\n+ [pyproject.toml](https://www.python.org/dev/peps/pep-0518/) support.\n+ [Check that all required plugins are installed](https://flakeheaven.readthedocs.io/en/latest/commands/missed.html).\n+ [Syntax highlighting in messages and code snippets](https://flakeheaven.readthedocs.io/en/latest/formatters.html#colored-with-source-code).\n+ [PyLint](https://github.com/PyCQA/pylint) integration.\n+ [Powerful GitLab support](https://flakeheaven.readthedocs.io/en/latest/formatters.html#gitlab).\n+ Codes management:\n    + Manage codes per plugin.\n    + Enable and disable plugins and codes by wildcard.\n    + [Show codes for installed plugins](https://flakeheaven.readthedocs.io/en/latest/commands/plugins.html).\n    + [Show all messages and codes for a plugin](https://flakeheaven.readthedocs.io/en/latest/commands/codes.html).\n    + Allow codes intersection for different plugins.\n\n![output example](./assets/grouped.png)\n\n## Compatibility\n\nFlakeHeaven supports all flake8 plugins, formatters, and configs. However, FlakeHeaven has it\'s own beautiful way to configure enabled plugins and codes. So, options like `--ignore` and `--select` unsupported. You can have flake8 and FlakeHeaven in one project if you want but enabled plugins should be explicitly specified.\n\n## Installation\n\n```bash\npython3 -m pip install --user flakeheaven\n```\n\n## Usage\n\nFirst of all, let\'s create `pyproject.toml` config:\n\n```toml\n[tool.flakeheaven]\n# optionally inherit from remote config (or local if you want)\nbase = "https://raw.githubusercontent.com/flakeheaven/flakeheaven/main/pyproject.toml"\n# specify any flake8 options. For example, exclude "example.py":\nexclude = ["example.py"]\n# make output nice\nformat = "grouped"\n# 80 chars aren\'t enough in 21 century\nmax_line_length = 90\n# show line of source code in output\nshow_source = true\n\n# list of plugins and rules for them\n[tool.flakeheaven.plugins]\n# include everything in pyflakes except F401\npyflakes = ["+*", "-F401"]\n# enable only codes from S100 to S199\nflake8-bandit = ["-*", "+S1??"]\n# enable everything that starts from `flake8-`\n"flake8-*" = ["+*"]\n# explicitly disable plugin\nflake8-docstrings = ["-*"]\n```\n\nShow plugins that aren\'t installed yet:\n\n```bash\nflakeheaven missed\n```\n\nShow installed plugins, used plugins, specified rules, codes prefixes:\n\n```bash\nflakeheaven plugins\n```\n\n![plugins command output](./assets/plugins.png)\n\nShow codes and messages for a specific plugin:\n\n```bash\nflakeheaven codes pyflakes\n```\n\n![codes command output](./assets/codes.png)\n\nRun flake8 against the code:\n\n```bash\nflakeheaven lint\n```\n\nThis command accepts all the same arguments as Flake8.\n\nRead [the documentation](https://flakeheaven.readthedocs.io/en/latest/) for more information.\n\n## Contributing\n\n1. Add tests when possible (eg for features / fixes / refactor, etc. )\n2. add your contribution to the code / docs\n3. Ensure your code passes all (both oririnal and your own) tests.\n4. commit using [proper header](https://www.conventionalcommits.org/en/v1.0.0/)\n5. create a PR\n\nContributions are welcome! A few ideas where you can contribute:\n\n+ Improve documentation.\n+ Add more tests.\n+ Improve performance.\n+ Found a bug? Fix it!\n+ Made an article about FlakeHeaven? Great! Let\'s add it into the `README.md`.\n+ Don\'t have time to code? No worries! Just tell your friends and subscribers about the project. More users -> more contributors -> more cool features.\n\nA convenient way to run tests is using [Poetry](https://python-poetry.org/docs/master/#installing-with-the-official-installer):\n\n```bash\ncurl -sSL https://install.python-poetry.org | python3 -\npoetry install\npoetry run pytest tests\n```\n\nThank you :heart:\n\n![](./assets/flaky.png)\n\nThe FlakeHeaven mascot (Flaky) is created by [@illustrator.way](https://www.instagram.com/illustrator.way/) and licensed under the [CC BY-SA 4.0](https://creativecommons.org/licenses/by-sa/4.0/) license.\n',
    'author': 'Gram',
    'author_email': 'master_fess@mail.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/flakeheaven/flakeheaven',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# built-in
import json
import os
from hashlib import md5
from pathlib import Path
from time import time
from typing import Optional

# external
from flake8.checker import FileChecker
from flake8.options.manager import OptionManager


CACHE_PATH = Path(os.environ.get('FLAKEHEAVEN_CACHE', Path.home() / '.cache' / 'flakeheaven'))
THRESHOLD = int(os.getenv('FLAKEHEAVEN_CACHE_TIMEOUT', 3600 * 24))  # default is 1 day


def prepare_cache(path=CACHE_PATH):
    if not path.exists():
        path.mkdir(parents=True)
        return
    for fpath in path.iterdir():
        if time() - fpath.stat().st_atime <= THRESHOLD:
            continue
        fpath.unlink()


class Snapshot:
    _exists: Optional[bool] = None
    _digest: Optional[str] = None
    _results = None

    def __init__(self, *, cache_path: Path, file_path: Path):
        self.cache_path = cache_path
        self.file_path = file_path

    @classmethod
    def create(cls, checker: FileChecker, options: OptionManager) -> 'Snapshot':
        hasher = md5()

        # plugins config
        plugins = json.dumps(options.plugins, sort_keys=True)
        hasher.update(plugins.encode())

        # file path
        file_path = Path(checker.filename).resolve()
        hasher.update(str(file_path).encode())

        return cls(
            cache_path=CACHE_PATH / (hasher.hexdigest() + '.json'),
            file_path=file_path,
        )

    def exists(self) -> bool:
        """Returns True if cache file exists and is actual.
        """
        if self._exists is not None:
            return self._exists

        if not self.cache_path.exists():
            self._exists = False
            return self._exists

        # digest is None for non-existent files (stdin)
        if self.digest is None:
            return False

        # check that file content wasn't changed since the snapshot
        cache = json.loads(self.cache_path.read_text())
        self._exists = self.digest == cache['digest']
        # if cache is valid results will be eventually requested.
        # let's save it for later use to avoid reading the cache twice
        if self._exists:
            self._results = cache['results']
        return self._exists  # type: ignore

    @property
    def digest(self) -> Optional[str]:
        """Get hex digest for the current content of the file
        """
        # we cache it because it requested twice: from `exists` and from `dumps`
        if self._digest is None:
            if not self.file_path.exists():
                return None
            hasher = md5()
            hasher.update(self.file_path.read_bytes())
            self._digest = hasher.hexdigest()
        return self._digest

    def dump(self, results) -> None:
        self.cache_path.write_text(self.dumps(results=results))

    def dumps(self, results) -> str:
        return json.dumps(dict(
            results=results,
            digest=self.digest,
        ))

    @property
    def results(self):
        """returns cached checks results for the given file
        """
        # results could be cached from `.exists()`.
        # however, we don't want to cache the results on requets
        # because they are always requested only once
        if self._results is not None:
            return self._results
        return json.loads(self.cache_path.read_text())['results']



# built-in
from hashlib import md5
from pathlib import Path


def make_baseline(path: Path, context: str, code: str, line: int) -> str:
    digest = md5()
    digest.update(path.as_posix().lstrip('./').encode())
    digest.update((context or str(line)).strip().encode())
    digest.update(code.encode())
    return digest.hexdigest()




# app
from ._pylint import PyLintChecker


# try:
#     from ._pylint import PyLintChecker
# except ImportError:
#     PyLintChecker = None


PLUGINS = dict(
    pylint=PyLintChecker,
)

__all__ = [
    'PLUGINS',
    'PyLintChecker',
]



# built-in
import atexit
import sys
import tempfile
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import List, Optional, Tuple

# app
from .._constants import NAME, VERSION, ExitCode
from .._patched import FlakeHeavenApplication
from .._types import CommandResult


FLAKE8_DEFAULTS = {
    'verbose': 0,
    'output_file': None,
}

EXCLUDED = {
    'jobs',
}


def _inject_flake8_defaults(parser: ArgumentParser) -> None:
    """Inject some flake8 defaults.

    Defaults taken from
    `flake8.main.options:register_preliminary_options`.

    Args:
        parser: The parser to inject defaults into.

    Returns:
        Mapping of option names to actions.
    """

    parser.add_argument(
        '-v',
        '--verbose',
        default=FLAKE8_DEFAULTS['verbose'],
        action='count',
        help='Print more information about what is happening in flake8.'
        ' This option is repeatable and will increase verbosity each '
        'time it is repeated.',
    )
    parser.add_argument(
        '--output-file',
        default=FLAKE8_DEFAULTS['output_file'],
        help='Redirect report to a file.',
    )


def _maybe_backup_old_log(local_args: Namespace) -> Tuple[Optional[str], Path]:
    bkp: Optional[str] = local_args.output_file

    temp_dir = tempfile.TemporaryDirectory(prefix='flakeheaven-')
    logfile = Path(temp_dir.name) / 'config.log'
    local_args.output_file = str(logfile)
    atexit.register(lambda: temp_dir.cleanup())
    return bkp, logfile


def _parse(sys_argv) -> Tuple[Namespace, List[str], Optional[str], Path]:
    parser = ArgumentParser(
        description=config_command.__doc__,
    )
    parser.add_argument(
        '--plugins-only',
        action='store_true',
        help='If set, show only the plugins section of the config',
    )
    parser.add_argument(
        '--flake8-logs',
        action='store_true',
        default=False,
        help='If set, also include flake8 logs',
    )

    # l1, argv = parser.parse_known_args(sys_argv)
    _inject_flake8_defaults(parser)
    local_args, argv = parser.parse_known_args(sys_argv)

    bkp, logfile = _maybe_backup_old_log(local_args)

    # fix default verbose offfset
    verbose = FLAKE8_DEFAULTS['verbose']
    if local_args.verbose != verbose:
        local_args.verbose -= verbose

    av = [
        '--output-file',
        local_args.output_file,
    ]
    if local_args.verbose:
        av.append(f'-{"v" * local_args.verbose}')
    argv.extend(av)

    return local_args, argv, bkp, logfile


def get_config(argv) -> Tuple[dict, Namespace, Path, Optional[str]]:
    local_args, argv, bkp, logfile = _parse(argv)
    app = FlakeHeavenApplication(program=NAME, version=VERSION)
    app.initialize(argv)

    app.options.output_file = bkp
    config = {
        k: v for k, v in vars(app.options).items()
        if k not in EXCLUDED
    }
    if local_args.plugins_only:
        config = config['plugins']
    return config, local_args, logfile, bkp


def config_command(argv) -> CommandResult:
    """Show flake8 configuration after flakeheaven consolidation."""

    config, local_args, logfile, bkp = get_config(argv)

    try:
        logs = (
            '\n'.join(
                line
                for line in logfile.read_text().splitlines()
                if local_args.flake8_logs or line.startswith('flake8.flakeheaven')
            ) + '\n'
        )
    except FileNotFoundError:
        logs = ''

    stdout = f'{config}\n'

    if bkp and bkp not in ('stdout', 'stderr'):
        fileobj = open(bkp, 'a+')
        atexit.register(lambda: fileobj.close())
        fileobj_logs = fileobj_app = fileobj
    else:
        fileobj_logs = getattr(sys, bkp or 'stderr')
        fileobj_app = getattr(sys, bkp or 'stdout')

    print(logs, file=fileobj_logs)
    print(stdout, file=fileobj_app)

    return ExitCode.OK, ''



# built-in
from collections import defaultdict
from typing import DefaultDict, Dict

# external
from flake8.style_guide import Violation

# app
from .._logic import color_code, color_description, colored
from ._colored import ColoredFormatter


class StatFormatter(ColoredFormatter):
    """
    Show count of every code occurance
    """
    _codes: DefaultDict[str, DefaultDict[str, int]]
    _msgs: DefaultDict[str, Dict[str, str]]

    def after_init(self) -> None:
        super().after_init()
        self._codes = defaultdict(lambda: defaultdict(int))
        self._msgs = defaultdict(dict)

    def format(self, error: Violation) -> None:
        plugin = getattr(error, 'plugin', '')
        self._codes[plugin][error.code] += 1
        self._msgs[plugin][error.code] = error.text

    def stop(self) -> None:
        for plugin, codes in sorted(self._codes.items()):
            self._write(colored(plugin, 'green'))
            for code, count in sorted(codes.items(), key=lambda x: x[-1], reverse=True):
                self._write('  {code:} | {count:>5} | {msg}'.format(
                    code=color_code(code.ljust(6)),
                    count=count,
                    msg=color_description(self._msgs[plugin][code]),
                ))



# built-in
from enum import IntEnum
from types import MappingProxyType

# app
from ._version import __version__


NAME = 'flakeheaven'
VERSION = __version__


# our own modern defaults
DEFAULTS = dict(
    # redefined defaults
    filename=['*.py', '*.ipynb', '*.md', '*.yml', '*.yaml', '*.rst', '*.rest'],
    format='colored',
    max_line_length=90,

    # flakeheaven options
    baseline=None,
    safe=False,
    plugins={
        'pyflakes': ['+*'],
        'pycodestyle': ['+*'],
    },
    exceptions={},

    # disabled by flakeheaven but required by flake8
    extend_exclude=[],
    ignore=[],
    extend_ignore=[],
    select=[],
    enable_extensions=[],
    per_file_ignores=[],
    statistics=False,
)


class ExitCode(IntEnum):
    OK = 0

    # CLI entrypoint
    NO_COMMAND = 1
    INVALID_COMMAND = 2

    # `installed` command
    NO_PLUGINS_INSTALLED = 11

    # `show` command
    NO_PLUGIN_NAME = 21
    IMPORT_ERROR = 22
    NO_CODES = 23

    TOO_MANY_ARGS = 31
    NOT_ENOUGH_ARGS = 32


# If a plugin isn't there, it still should be supported.
# However, support for these plugins is tested on CI.
KNOWN_PLUGINS = MappingProxyType({
    'dlint': 'https://github.com/dlint-py/dlint',
    'flake8-2020': 'https://github.com/asottile/flake8-2020',
    'flake8-alfred': 'https://github.com/datatheorem/flake8-alfred',
    'flake8-annotations-complexity': 'https://github.com/best-doctor/flake8-annotations-complexity',
    'flake8-bandit': 'https://github.com/tylerwince/flake8-bandit',
    'flake8-black': 'https://github.com/peterjc/flake8-black',
    'flake8-broken-line': 'https://github.com/sobolevn/flake8-broken-line',
    'flake8-builtins': 'https://github.com/gforcada/flake8-builtins',
    'flake8-coding': 'https://github.com/tk0miya/flake8-coding',
    'flake8-cognitive-complexity': 'https://github.com/Melevir/flake8-cognitive-complexity',
    'flake8-comprehensions': 'https://github.com/adamchainz/flake8-comprehensions',
    'flake8-debugger': 'https://github.com/JBKahn/flake8-debugger',
    'flake8-docstrings': 'https://gitlab.com/pycqa/flake8-docstrings',  # pydocstyle
    'flake8-eradicate': 'https://github.com/sobolevn/flake8-eradicate',
    'flake8-executable': 'https://github.com/xuhdev/flake8-executable',
    'flake8-expression-complexity': 'https://github.com/best-doctor/flake8-expression-complexity',
    'flake8-fixme': 'https://github.com/tommilligan/flake8-fixme',
    'flake8-functions': 'https://github.com/best-doctor/flake8-functions',
    'flake8-logging-format': 'https://github.com/globality-corp/flake8-logging-format',
    'flake8-mutable': 'https://github.com/ebeweber/flake8-mutable',
    'flake8-mypy': 'https://github.com/ambv/flake8-mypy',
    'flake8-pep3101': 'https://github.com/gforcada/flake8-pep3101',
    'flake8-pie': 'https://github.com/sbdchd/flake8-pie',
    'flake8-print': 'https://github.com/JBKahn/flake8-print',
    'flake8-printf-formatting': 'https://github.com/atugushev/flake8-printf-formatting',
    'flake8-pyi': 'https://github.com/ambv/flake8-pyi',
    'flake8-quotes': 'https://github.com/zheller/flake8-quotes',
    'flake8-requirements': 'https://github.com/Arkq/flake8-requirements',
    'flake8-rst-docstrings': 'https://github.com/peterjc/flake8-rst-docstrings',
    'flake8-spellcheck': 'https://github.com/MichaelAquilina/flake8-spellcheck',
    'flake8-sql': 'https://github.com/pgjones/flake8-sql',
    'flake8-string-format': 'https://github.com/xZise/flake8-string-format',
    'flake8-todo': 'https://github.com/schlamar/flake8-todo',
    'flake8-use-fstring': 'https://github.com/MichaelKim0407/flake8-use-fstring',
    'flake8-variables-names': 'https://github.com/best-doctor/flake8-variables-names',

    # framework-specific
    'flake8-django': 'https://github.com/rocioar/flake8-django',
    'flake8-scrapy': 'https://github.com/stummjr/flake8-scrapy',
    'pandas-vet': 'https://github.com/deppen8/pandas-vet',

    # tests
    'flake8-aaa': 'https://github.com/jamescooke/flake8-aaa',
    'flake8-mock': 'https://github.com/aleGpereira/flake8-mock',
    'flake8-pytest': 'https://github.com/vikingco/flake8-pytest',
    'flake8-pytest-style': 'https://github.com/m-burst/flake8-pytest-style',

    # PyCQA
    'flake8-bugbear': 'https://github.com/PyCQA/flake8-bugbear',
    'flake8-commas': 'https://github.com/PyCQA/flake8-commas',
    'mccabe': 'https://github.com/PyCQA/mccabe',
    'pep8-naming': 'https://github.com/PyCQA/pep8-naming',
    'pylint': 'https://github.com/PyCQA/pylint',

    # imports
    'flake8-future-import': 'https://github.com/xZise/flake8-future-import',
    'flake8-import-order': 'https://github.com/PyCQA/flake8-import-order',
    'flake8-isort': 'https://github.com/gforcada/flake8-isort',
    'flake8-absolute-import': 'https://github.com/bskinn/flake8-absolute-import',
    'flake8-tidy-imports': 'https://github.com/adamchainz/flake8-tidy-imports',

    # built-in in flake8
    'pycodestyle': 'https://github.com/PyCQA/pycodestyle',
    'pyflakes': 'https://github.com/PyCQA/pyflakes',
})



# built-in
import re
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

# external
from flake8.utils import fnmatch


REX_NAME = re.compile(r'[-_.]+')
ALIASES = {
    'aaa': 'flake8-aaa',
    'flake-mutable': 'flake8-mutable',
    'flake8-pandas-vet': 'pandas-vet',
    'import-order': 'flake8-import-order',
    'logging-format': 'flake8-logging-format',
    'naming': 'pep8-naming',
    'pyflakes': 'pyflakes',
    'sql': 'flake8-sql',
    'use-fstring-format': 'flake8-use-fstring',
    'use-fstring-percent': 'flake8-use-fstring',
    'use-fstring-prefix': 'flake8-use-fstring',
}
PluginsType = Dict[str, List[str]]


def get_plugin_name(plugin: Dict[str, Any]) -> str:
    """Get plugin name from plugin info

    Users expect the same plugin name as the name of the package that provides plugin.
    However, that's not true for some plugins.
    Also, some plugins has different module name, that doesn't match to package.

    Lookup order:

    1. Ad-hoc aliases when nothing match
    2. Normalized name that starts with `flake8`
    3. Normalized name that starts with `pep`
    4. `plugin_name`
    """
    if not plugin:
        return 'UNKNOWN'
    if plugin['plugin_name'] in ALIASES:
        return ALIASES[plugin['plugin_name']]

    names = (plugin['plugin_name'], plugin['plugin'].__module__)
    names = [REX_NAME.sub('-', name).lower() for name in names]
    for name in names:
        if name.startswith('flake8'):
            return name
    for name in names:
        if name.startswith('pep8'):
            return name
    return names[0]


def get_plugin_rules(plugin_name: str, plugins: PluginsType) -> List[str]:
    """Get rules for plugin from `plugins` in the config

    Plugin name can be specified as a glob expression.
    So, it's not trivial to match the right one

    1. Try to find exact match (normalizing ass all packages names normalized)
    2. Try to find globs that match and select the longest one (nginx-style)
    3. Return empty list if nothing is found.
    """
    if not plugins:
        return []
    plugin_name = REX_NAME.sub('-', plugin_name).lower()
    # try to find exact match
    for pattern, rules in plugins.items():
        if '*' not in pattern and REX_NAME.sub('-', pattern).lower() == plugin_name:
            return rules

    # try to find match by pattern and select the longest
    best_match = (0, [])  # type: Tuple[int, List[str]]
    for pattern, rules in plugins.items():
        if not fnmatch(filename=plugin_name, patterns=[pattern]):
            continue
        match = len(pattern)
        if match > best_match[0]:
            best_match = match, rules
    if best_match[0]:
        return best_match[1]

    return []


def check_include(code: str, rules: List[str]) -> bool:
    """
    0. Validate rules

    1. Return True if rule explicitly included
    2. Return False if rule explicitly excluded

    3. Return True if the latest glob-matching rule is include
    4. Return False if the latest glob-matching rule is exclude
    """
    # always report exceptions in file processing
    if code in ('E902', 'E999'):
        return True

    for rule in rules:
        if len(rule) < 2 or rule[0] not in {'-', '+'}:
            raise ValueError('invalid rule: `{}`'.format(rule))

    for rule in reversed(rules):
        if code.lower() == rule[1:].lower():
            return rule[0] == '+'

    include = False
    for rule in rules:
        if fnmatch(code, patterns=[rule[1:]]):
            include = rule[0] == '+'
    return include


def get_exceptions(
    path: Union[str, Path], exceptions: Dict[str, PluginsType], root: Path = None,
) -> PluginsType:
    """
    Just like in get_plugin_rules. The algorithm:

    1. Try to find exact match (normalizing as all packages names normalized)
    2. Try to find globs that match and select the longest one (nginx-style)
    """
    if not exceptions:
        return dict()
    if isinstance(path, str):
        path = Path(path)
    if root is None:
        root = Path().resolve()
    try:
        path = path.resolve().relative_to(root).as_posix()
    except ValueError:
        # path is not in the project root
        return dict()
    exceptions = sorted(
        exceptions.items(),
        key=lambda item: len(item[0]),
    )

    aggregated_rules = dict()

    # glob
    for path_rule, rules in exceptions:
        if '*' not in path_rule:
            continue
        if fnmatch(filename=path, patterns=[path_rule]):
            aggregated_rules.update(rules)

    # prefix
    for path_rule, rules in exceptions:
        if '*' in path_rule:
            continue
        if path.startswith(path_rule):
            aggregated_rules.update(rules)

    return aggregated_rules
